@@ -18,339 +18,12 @@
 
 **************************************************************************/
 
-/*
- * This software package has been modified by the Worldforge Project 
- * in the following ways:
+/**
+ * This software package has been extensively modified by members of the
+ * Worldforge Project. See the file ChangeLog for details.
  *
- * $Log: skstream.cpp,v $
- * Revision 1.44  2003/09/26 22:26:44  alriddoch
- *  2003-09-26 Al Riddoch <alriddoch@zepler.org>
- *     - Add option to get streams remote details as reverse lookup, rather
- *       than just presentation form.
+ * $Id: skstream.cpp,v 1.53 2005/09/15 10:38:56 alriddoch Exp $
  *
- * Revision 1.43  2003/09/26 14:38:43  alriddoch
- *  2003-09-26 Al Riddoch <alriddoch@zepler.org>
- *     - Write some tests to pick up the socket and name resolver libs on
- *       System V.
- *     - Clean up handling of libs required to get socket calls in general.
- *     - ping/ping.cpp, skstream/skserver.cpp: Add missing stdio include.
- *     - skstream/skserver.h: Remove non-required string.h include.
- *     - skstream/skstream.cpp: Add mising stdio include, and clean up sprintf()
- *       usage.
- *
- * Revision 1.42  2003/09/26 10:49:03  alriddoch
- *  2003-09-26 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.cpp: Add code that handles converting service to
- *       presentation format in a protocol independant way.
- *
- * Revision 1.41  2003/09/26 01:29:41  alriddoch
- *  2003-09-26 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.h: Add accessors for address size members.
- *     - ping/ping.cpp: Clean up use of libc and system calls.
- *     - skstream/skstream.cpp: Clean up handling socket address sizes
- *       when setting socket by descriptor in the buffer objects.
- *       Add code that handles converting address to presentation format
- *       in a protocol independant way.
- *
- * Revision 1.40  2003/09/25 17:05:49  alriddoch
- *  2003-09-25 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skpoll.cpp, skstream/skserver.cpp, skstream/skstream.cpp:
- *       Finish off fully qualifying all libc and system calls as being
- *       explicitly in the global namespace.
- *
- * Revision 1.39  2003/09/25 16:39:50  alriddoch
- *  2003-09-25 Al Riddoch <alriddoch@zepler.org>
- *     - Remove messy attempt to handle IPv6 addresses without getaddrinfo
- *
- * Revision 1.38  2003/09/25 01:17:10  alriddoch
- *  2003-09-25 James Turner <james@worldforge.org>
- *     - skstream/skstream.cpp: Fully qualify all libc and system calls as
- *       being explicitly in the global namespace.
- *
- * Revision 1.37  2003/09/24 21:02:11  alriddoch
- *  2003-09-24 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.cpp: Implement non-blocking connect with addrinfo list.
- *
- * Revision 1.36  2003/09/24 16:20:03  alriddoch
- *  2003-09-24 Al Riddoch <alriddoch@zepler.org>
- *     - Typo in in_addr_t check fixed.
- *     - skstream/skstream.cpp: Missed out typedef keyword fix.
- *
- * Revision 1.35  2003/09/24 14:47:06  alriddoch
- *  2003-09-24 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.cpp: Add check, and fallback definition for
- *       in_addr_t.
- *
- * Revision 1.34  2003/09/24 01:05:46  alriddoch
- *  2003-09-24 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.h, skstream/skstream.cpp: Re-write connecting
- *       code for IP datagram and stream sockets so they try each of the
- *       results from getaddrinfo in turn. Needs work on non-blocking
- *       sockets.
- *
- * Revision 1.33  2003/08/25 17:18:27  alriddoch
- *  2003-08-23 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.h, skstream/skstream.cpp: Add extra argument
- *       to dgram_streambuf::setTarget() so protocol can be passed in.
- *       Add #warnings if the old non getaddrinfo code is being used.
- *
- * Revision 1.32  2003/08/23 22:13:55  alriddoch
- *  2003-08-23 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstreamconfig.h.in, skstream/skstream_unix.h,
- *       skstream/skstream.h, skstream/skstream.cpp,
- *       skstream/skserver_unix.h, skstream/skserver.h, skstream/skserver.cpp:
- *       Remove as much platform sensitive code from the headers as possible,
- *       and simplify the generated header. Improve formatting and header
- *       guards for readability.
- *     - skstream/skserver_unix.h: Add a new constructor for unix_socket_server
- *       so it can be instanced without already knowing the details of the
- *       service.
- *
- * Revision 1.31  2003/08/23 20:19:46  alriddoch
- *  2003-08-23 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.h, skstream/skstream.cpp: Modify udp_socket_stream
- *       constructor so that it does not create the socket immediatly.
- *     - skstream/skstream.cpp: Re-write dgram_socketbuf::setTarget() so it
- *       uses getaddrinfo to create the socket, and get the address in
- *       a protocol independant way.
- *
- * Revision 1.30  2003/08/23 01:14:44  alriddoch
- *  2003-08-23 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.cpp: Use getaddrinfo to handle the port number.
- *
- * Revision 1.29  2003/08/23 00:37:38  alriddoch
- * *** empty log message ***
- *
- * Revision 1.28  2003/08/08 23:56:26  alriddoch
- *  2003-08-08 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.cpp, skstream/skstream_unix.h: Include skstream
- *       header with its fully qualified name for compatability, and move
- *       the unix system header out of the header into the cpp file.
- *     - skstream/skserver.cpp, skstream/skserver.h: Add new base class for ip
- *       socket types, and put in the ip address related functionality into it,
- *       so basic class can be used as a base for unix socket class.  Move
- *       can_accept() into the base class as it applies to any listen socket.
- *     - skstream/skserver_unix.h, skstream/skserver.cpp: Add a class for
- *       unix listen sockets.
- *
- * Revision 1.27  2003/07/30 23:17:55  alriddoch
- *  2003-07-30 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skserver.cpp, skstream/skserver.h, skstream/skstream.cpp,
- *       skstream/skstream.h, skstream/skstream_unix.h: Move virtual
- *       function implementations into .cpp files.
- *
- * Revision 1.26  2003/07/23 17:00:29  alriddoch
- *  2003-07-23 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstreamconfig.h.pbx, skstream/skstreamconfig.h.in,
- *       skstream/skstreamconfig.h.windows: Removed some defines which
- *       are the same on all platforms, and put them in the main header.
- *       Avoid streambuf header as it varies on some systems, and rely
- *       on iostream instead.
- *     - Remove the streambuf header test.
- *     - skstream/skstream.h: Remove the old platform specific code,
- *       take on some changes from the skstreamconfig header, and clean
- *       up some comments.
- *     - skstream/skstream.cpp: Re-write the resolver code for platforms
- *       without getaddrinfo() so it is cleaner, and covers more cases.
- *
- * Revision 1.25  2003/05/06 21:53:11  alriddoch
- *  2003-05-06 Al Riddoch <alriddoch@zepler.org>
- *     - skstream/skstream.h, skstream/skstream.cpp, skstream_unix.h:
- *       Re-work basic_socket_stream so it can have either stream or datagram
- *       buffers.
- *     - ping/ping.cpp, ping/ping.h, test/basicskstreamtest.h,
- *       test/childskstreamtest.h, test/skservertest.h: Get the tests and examples
- *       building again.
- *
- * Revision 1.24  2003/05/04 01:14:07  alriddoch
- *  2003-05-04 Al Riddoch <alriddoch@zepler.org>,
- *     - skstream/skstream.cpp: Add a check to detect IPv6 sockaddr in
- *       headers, and treat address explicitly as an IPv6 address if
- *       it is.
- *
- * Revision 1.23  2003/05/04 00:34:30  alriddoch
- *  2003-05-04 Al Riddoch <alriddoch@zepler.org>,
- *     - Add a second pkgconfig file for apps that need unix socket support.
- *     - Rename sksystem.h as skstreamconfig.h, and put it in an architecture
- *       dependent subdir. Modify pkgconfig files to handle this.
- *     - Handle errors in a more general way to reduce the number
- *       of places where pre-processor conditionals are required.
- *     - Remove some redundant code for dealing with win32 that has
- *       now been superceded.
- *
- * Revision 1.22  2003/04/22 15:14:59  alriddoch
- *  2003-04-22 Al Riddoch <alriddoch@zepler.org>,
- *     - Add configure checks and support for protocol family independant
- *       socket address functions, so we can support IPv6 as well as IPv4.
- *
- * Revision 1.21  2003/04/16 14:02:35  alriddoch
- *  2003-04-16 Al Riddoch <alriddoch@zepler.org>,
- *     - Fix up includes so they work properlly
- *
- * Revision 1.20  2003/03/18 22:46:45  alriddoch
- *  2003-03-18 Al Riddoch <alriddoch@zepler.org>,
- *     - Improve safety of some constructors, and move IP (host and port)
- *       related functions into apropriate classes.
- *     - Add in unix sockets, based on a configure check.
- *
- * Revision 1.19  2003/03/16 12:27:40  alriddoch
- *  2003-03-16 Al Riddoch <alriddoch@zepler.org>,
- *     - skstream.cpp: Fix problem with virtual method being called
- *       after destructor.
- *
- * Revision 1.18  2003/03/14 19:33:11  alriddoch
- *  2003-03-14 Al Riddoch <alriddoch@zepler.org>,
- *     - skstream.h, skstream.cpp: Re-work sockbuf class so it is
- *       not specific to any one type of socket. Remove inet
- *       functionality from base classes. Derive new buffer
- *       classes for stream and datagram sockets, and use
- *       stream version in current socket classes. This should
- *       be considered work in progress. Do not port your code.
- *
- * Revision 1.17  2002/12/09 22:13:21  rsteinke
- *     - created basic_socket, a virtual base class
- *       for basic_socket_stream and basic_socket_server,
- *       so that the polling code has a common base
- *       with getSocket() to access
- *     - created basic_socket_poll polling class
- *     - changed tcp_socket_stream::is_ready() to isReady()
- *       to match standards for function naming, left is_ready()
- *       as a wrapper of isReady() for compatibility
- *
- * Revision 1.16  2002/11/04 22:49:15  alriddoch
- *  2002-11-04 Al Riddoch <alriddoch@zepler.org>,
- *     - Add doxygen docs.
- *
- * Revision 1.15  2002/10/14 05:05:45  malcolm
- * Added a Sleep(0) call during check for getsockopt() that appears to be needed
- * to circumvent a bug during nonblocking connect.  See MS KB Q165989.
- *
- * Revision 1.14  2002/07/22 19:33:48  rsteinke
- *     - changed sizeof(sockaddr) to sizeof(sockaddr_in) in
- *       several locations, so that we pass the correct size
- *       of in_peer and out_peer in socketbuf
- *     - reordered initializers in socketbuf constructors
- *       to match their declaration order in the class
- *
- * Revision 1.13  2002/07/16 21:10:06  malcolm
- * Had to add another tiny tweak for win32 nonblocking stuff.  ioctlsocket()
- * explicitly requires a pointer to an unsigned long, not just a 0 (NULL),
- * for the third argument, in windows.  Sheesh.
- *
- * Revision 1.12  2002/07/16 18:14:25  grimicus
- * 2002-07-16 Dan Tomalesky <grim@xynesis.com>
- *     * skstream.cpp: added a small fix to a bug with is_ready.  wasn't
- *       compiling on linux and seems code was wrong in the select because it was
- *       doing |= to the select() instead of != (i.e., if it doesn't return 1,
- *       there is no ready socket or an error)
- *
- * Revision 1.11  2002/07/16 01:12:19  malcolm
- * Fixed call for win32 cases with is_ready() - it was failing before with
- * WSAEFAULT - bad pointer.  Works now.  Tested with uclient; all is good.
- *
- * Revision 1.10  2002/07/15 21:21:17  alriddoch
- *  07/15/2002 Al Riddoch <alriddoch@zepler.org>,
- *     - skstream.cpp: Handle nonblocking connect in win32
- *     - skstream.h: Include correct streambuf header on Linux.
- *
- * Revision 1.9  2002/07/04 10:23:56  jmt
- *  07/03/2002 James Turner <james@worldforge.org>
- * 	-Added configuration case for Darwin / OS-X
- * 	-Fixed Rsteinke's changes to use typedefs for socklen / errnum
- *
- * Revision 1.8  2002/06/12 01:21:47  rsteinke
- *     -Added an optional "milliseconds" argument to
- *      tcp_socket_stream::is_ready(), to take advantage
- *      of the timeout in select()
- *
- * Revision 1.7  2002/06/12 00:27:40  rsteinke
- *     -Fixed many bugs in nonblocking tcp socket connect code
- *
- * Revision 1.6  2002/05/21 07:29:36  malcolm
- * Added rsteinke's nonblocking connect patch.  Works on linux; does not break API
- * (I bumped version to 0.2.3 anyway).  May not work on win32, though I did test it
- * and socket communication does happen.
- *
- * Revision 1.5  2002/04/08 20:02:00  xmp
- * Just a few fixes to MSVC support, removing a few unnessesary std::'s and shifting a default argument to a header.
- *
- * Revision 1.4  2002/04/08 19:47:12  malcolm
- * Changes to allow clean compilation on Microsoft Visual C++.  I must go and
- * boil my keyboard now.  Constructor code for basic_socket_stream moved out of
- * header and into .cpp.
- *
- * MS Project files are untested (I edited them!  Oh no!)
- *
- * Revision 1.3  2002/02/26 20:33:55  grimicus
- * 02/26/2002 Dan Tomalesky <grim@xynesis.com>
- *     * Added test cases for skserver and friends
- *
- *     * Adding .cvsignore files so that it doesn't mess with non-cvs files anymore
- *
- *     * In skserver.cpp and skstream.cpp, in the close() methods, I commented out
- *       the return; in the error part of the shutdown() call because it is
- *       possible that a socket can exist without it actually having been used,
- *       so this could error out on those cases and the socket is never actually
- *       closed.
- *
- *     * In skserver.h, added can_accept() to tcp_socket_server so that it can be
- *       checked to see if the server socket has any connections available, so that
- *       accept() can then be called. (if it returns false, if accept is called,
- *       it will block until a connection is made)
- *
- *     * Removed the #include <iostream> from skserver.h and skstream.h as they are
- *       not actually needed for any of the code. (else it comes in from some other
- *       include, I'm not positive)
- *
- *     * Made some formatting changes in skserver.h along the same lines as I have
- *       been doing throughout the code.
- *
- *     * Added testClose() to basicskstreamtest.
- *
- *     * Changed the socket created in basicskstreamtest from SOCK_STREAM to
- *       SOCK_DGRAM though it doesn't make any difference what so ever in the
- *       testing.
- *
- *     * Added the skservertests into the test runner.
- *
- * Revision 1.2  2002/02/24 03:15:41  grimicus
- * 02/23/2002 Dan Tomalesky <grim@xynesis.com>
- *
- *     * Added in CVS logging variable so that changes show up in modified files
- *       This will help keep changes made by worldforge in each file that is
- *       changed as required by the GPL.
- *
- *     * Changed some parameter variables to have better meaning.
- *       (ad -> address, etc.)
- *
- *     * Added some code into tcp_sk_stream::open so that it calls setLastError()
- *       when the connection fails.
- *
- *     * Added some comments into skstream.h to better describe SOCKET_TYPE as
- *       there can be some confusion between what it is actually for
- *       (pointer/file descriptor/windows cludge of the socket) and the various
- *       types of sockets (tcp, udp, raw, etc)
- *
- *     * Changed some more formatting for readability.
- *
- *     * Uncommented some commented out code in skstream.h so that the sync()
- *       method returns 0 on the else, rather than not returning anything.
- *
- *     * Added some code into setBroadcast() so that setLastError() is called
- *       if it fails to perform the getsocketopt().
- *
- *     * Modified the test/Makefile.am to remove the header files from the SOURCES
- *       as the .h files do not seem to affect the build.
- *
- *     * Updated all the current test so that they use a socket instead of the
- *       absolutely wrong stuff I was doing before.
- *
- *     * Added tests for tcp, udp, and raw skstreams child classes.
- *
- * Revision 1.1  2002/01/07 23:02:08  rykard
- * Adding the new version of skstream/FreeSockets to cvs.  
- * Note there are some API changes and new features in this version, so I 
- * didn't just commit over the older one.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -370,6 +43,12 @@
 #include <cstdio>
 #include <cassert>
 
+#ifdef _WIN32
+#define SHUT_RD SD_RECEIVE
+#define SHUT_WR SD_SEND
+#define SHUT_RDWR SD_BOTH
+#endif
+
 static inline int getSystemError()
 {
   #ifdef _WIN32
@@ -409,15 +88,12 @@ socketbuf::socketbuf(SOCKET_TYPE sock, unsigned insize, unsigned outsize)
   char* buffer = new char[bufsize];
   ::memset(buffer,0,bufsize);
   // Setup the buffer
-  if(this != setbuf(buffer, bufsize)) {
-    _buffer = NULL;
-  }
+  setbuf(buffer, bufsize);
+
   _timeout.tv_sec  = 0;
   _timeout.tv_usec = 0;
 
-  ::getpeername(sock,(sockaddr*)&out_peer, &out_p_size);
-  in_peer = out_peer;
-  in_p_size = out_p_size;
+  setSocket(sock);
 }
 
 // Constructor
@@ -427,15 +103,12 @@ socketbuf::socketbuf(SOCKET_TYPE sock, char* buf, int length)
       Timeout(false)
 {
   _buffer = NULL;
-  if(this != setbuf(buf,length)) {
-    _buffer = NULL;
-  }
+  setbuf(buf, length);
+
   _timeout.tv_sec  = 0;
   _timeout.tv_usec = 0;
 
-  ::getpeername(sock,(sockaddr*)&out_peer, &out_p_size);
-  in_peer = out_peer;
-  in_p_size = out_p_size;
+  setSocket(sock);
 }
 
 // Destructor
@@ -447,10 +120,12 @@ socketbuf::~socketbuf(){
 void socketbuf::setSocket(SOCKET_TYPE sock)
 {
   _socket = sock;
-  out_p_size = sizeof(sockaddr);
-  ::getpeername(sock, (sockaddr*)&out_peer, &out_p_size);
-  in_peer = out_peer;
-  in_p_size = out_p_size;
+  out_p_size = sizeof(out_peer);
+  if(sock != INVALID_SOCKET) {
+    ::getpeername(sock, (sockaddr*)&out_peer, &out_p_size);
+    in_peer = out_peer;
+    in_p_size = out_p_size;
+  }
 }
 
 
@@ -468,13 +143,15 @@ int socketbuf::sync()
     }
 }
 
-std::streambuf* socketbuf::setbuf(char* buf, long len)
+std::streambuf * socketbuf::setbuf(std::streambuf::char_type * buf,
+                                   std::streamsize len)
 {
     if((buf != NULL) && (len > 0)) {
       _buffer = buf;
       setp(_buffer, _buffer+(len >> 1));
       setg(_buffer+(len >> 1), _buffer+len, _buffer+len);
     }
+
     return this;
 }
 
@@ -815,6 +492,14 @@ int dgram_socketbuf::underflow() {
   return (int)(unsigned char)(*gptr()); // traits::not_eof(...)
 }
 
+/////////////////////////////////////////////////////////////////////////////
+// class basic_socket implementation
+/////////////////////////////////////////////////////////////////////////////
+
+basic_socket::~basic_socket()
+{
+}
+
 /////////////////////////////////////////////////////////////////////////////
 // class basic_socket_stream implementation
 /////////////////////////////////////////////////////////////////////////////
@@ -823,7 +508,7 @@ int dgram_socketbuf::underflow() {
 
 basic_socket_stream::basic_socket_stream(socketbuf & buffer,
                                                  int proto)
-    : std::iostream(&_sockbuf), _sockbuf(buffer),
+    : std::iostream(&buffer), _sockbuf(buffer),
       protocol(proto), LastError(0)
 {
   startup();
@@ -832,45 +517,37 @@ basic_socket_stream::basic_socket_stream(socketbuf & buffer,
 
 basic_socket_stream::~basic_socket_stream()
 {
-    shutdown();
-}
-
-bool basic_socket_stream::operator!()
-{
-    return fail();
-}
-
-bool basic_socket_stream::timeout() const
-{
-    return _sockbuf.timeout();
+  if(is_open()) {
+    ::shutdown(_sockbuf.getSocket(), SHUT_RDWR);
+    ::closesocket(_sockbuf.getSocket());
+  }
+  delete &_sockbuf;
 }
 
 SOCKET_TYPE basic_socket_stream::getSocket() const
 {
-    return _sockbuf.getSocket();
+  return _sockbuf.getSocket();
 }
 
 // System dependant initialization
 bool basic_socket_stream::startup() {
-  #ifdef _WIN32
-    const unsigned wMinVer = 0x0101;	// request WinSock v1.1 (at least)
-    WSADATA wsaData;
-    LastError = WSAStartup(wMinVer, &wsaData);
-    return (LastError == 0);
-  #else // _WIN32
-    return true;
-  #endif // _WIN32
+#ifdef _WIN32
+  const unsigned wMinVer = 0x0101;	// request WinSock v1.1 (at least)
+  WSADATA wsaData;
+  LastError = WSAStartup(wMinVer, &wsaData);
+  return (LastError == 0);
+#else // _WIN32
+  return true;
+#endif // _WIN32
 }
 
 // System dependant finalization
-bool basic_socket_stream::shutdown() {
-  if(is_open()) close();
-  #ifdef _WIN32
-    LastError = WSACleanup();
-    return (LastError == 0);
-  #else // _WIN32
-    return true;
-  #endif // _WIN32
+void basic_socket_stream::shutdown() {
+  if(is_open()) {
+    if(::shutdown(_sockbuf.getSocket(), SHUT_RDWR) == SOCKET_ERROR) {
+      setLastError();
+    }
+  }
 }
 
 // private function that sets the internal variable LastError
@@ -881,12 +558,11 @@ void basic_socket_stream::setLastError() const {
 // closes a socket connection
 void basic_socket_stream::close() {
   if(is_open()) {
-    if(::shutdown(_sockbuf.getSocket(),0) == SOCKET_ERROR) {
+    if(::shutdown(_sockbuf.getSocket(), SHUT_RDWR) == SOCKET_ERROR) {
       setLastError();
       //not necessarily a returning offense because there could be a socket
       //open that has never connected to anything and hence, does not need
       //to be shutdown.
-      //return;
     }
 
     if(::closesocket(_sockbuf.getSocket()) == SOCKET_ERROR) {
@@ -924,9 +600,49 @@ bool basic_socket_stream::setBroadcast(bool opt)
 // class tcp_socket_stream implementation
 /////////////////////////////////////////////////////////////////////////////
 
+tcp_socket_stream::tcp_socket_stream() :
+                  basic_socket_stream(*new stream_socketbuf(INVALID_SOCKET)),
+                  _connecting_socket(INVALID_SOCKET),
+                  _connecting_address(0),
+                  _connecting_addrlist(0),
+                  stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  protocol = FreeSockets::proto_TCP;
+}
+
+tcp_socket_stream::tcp_socket_stream(SOCKET_TYPE socket) :
+                  basic_socket_stream(*new stream_socketbuf(socket)),
+                  _connecting_socket(INVALID_SOCKET),
+                  _connecting_address(0),
+                  _connecting_addrlist(0),
+                  stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  protocol = FreeSockets::proto_TCP;
+}
+
+tcp_socket_stream::tcp_socket_stream(const std::string& address, int service,
+                                     bool nonblock) :
+                  basic_socket_stream(*new stream_socketbuf(INVALID_SOCKET)),
+                  _connecting_socket(INVALID_SOCKET),
+                  _connecting_address(0),
+                  _connecting_addrlist(0),
+                  stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  protocol = FreeSockets::proto_TCP;
+  open(address, service, nonblock);
+}
+
+tcp_socket_stream::tcp_socket_stream(const std::string& address, int service,
+                                     unsigned int milliseconds) :
+                  basic_socket_stream(*new stream_socketbuf(INVALID_SOCKET)),
+                  _connecting_socket(INVALID_SOCKET),
+                  _connecting_address(0),
+                  _connecting_addrlist(0),
+                  stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  protocol = FreeSockets::proto_TCP;
+  open(address, service, milliseconds);
+}
+
 tcp_socket_stream::~tcp_socket_stream() { 
-  shutdown(); 
   if(_connecting_socket != INVALID_SOCKET) {
+    ::shutdown(_connecting_socket, SHUT_RDWR);
     ::closesocket(_connecting_socket);
 #ifdef HAVE_GETADDRINFO
     ::freeaddrinfo(_connecting_addrlist);
@@ -1120,6 +836,16 @@ void tcp_socket_stream::open(const std::string & address,
   _sockbuf.setSocket(_socket);
 }
 
+void tcp_socket_stream::open(const std::string & address, int service,
+                             unsigned int milliseconds)
+{
+  open(address, service, true);
+  if(!isReady(milliseconds)) {
+    close();
+    fail();
+  }
+}
+
 void tcp_socket_stream::close()
 {
   if(_connecting_socket != INVALID_SOCKET) {
@@ -1292,15 +1018,17 @@ bool tcp_socket_stream::isReady(unsigned int milliseconds)
 // class udp_socket_stream implementation
 /////////////////////////////////////////////////////////////////////////////
 
-udp_socket_stream::udp_socket_stream() : basic_socket_stream(dgram_sockbuf),
-                                         dgram_sockbuf(INVALID_SOCKET)
+udp_socket_stream::udp_socket_stream() :
+                  basic_socket_stream(*new dgram_socketbuf(INVALID_SOCKET)),
+                  dgram_sockbuf((dgram_socketbuf&)_sockbuf)
 {
   protocol = FreeSockets::proto_UDP; 
 }
 
 udp_socket_stream::~udp_socket_stream()
 {
-    shutdown();
+  // Don't close the main socket, that is done in the basic_socket_stream
+  // destructor
 }
 
 #ifdef SKSTREAM_UNIX_SOCKETS
@@ -1313,11 +1041,36 @@ udp_socket_stream::~udp_socket_stream()
 // class unix_socket_stream implementation
 /////////////////////////////////////////////////////////////////////////////
 
-unix_socket_stream::~unix_socket_stream() { 
-    shutdown(); 
-    if (_connecting_socket != INVALID_SOCKET) {
-        ::close(_connecting_socket);
-    }
+unix_socket_stream::unix_socket_stream() :
+                   basic_socket_stream(*new stream_socketbuf(INVALID_SOCKET)),
+                   _connecting_socket(INVALID_SOCKET),
+                   stream_sockbuf((stream_socketbuf&)_sockbuf) {
+}
+
+unix_socket_stream::unix_socket_stream(const std::string & address,
+                                       bool nonblock) :
+                   basic_socket_stream(*new stream_socketbuf(INVALID_SOCKET)),
+                   _connecting_socket(INVALID_SOCKET),
+                   stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  open(address, nonblock);
+}
+
+unix_socket_stream::unix_socket_stream(const std::string & address,
+                                       unsigned int milliseconds) :
+                   basic_socket_stream(stream_sockbuf),
+                   _connecting_socket(INVALID_SOCKET),
+                   stream_sockbuf((stream_socketbuf&)_sockbuf) {
+  open(address, milliseconds);
+}
+
+unix_socket_stream::~unix_socket_stream()
+{ 
+  // Don't close the main socket, that is done in the basic_socket_stream
+  // destructor
+  if(_connecting_socket != INVALID_SOCKET) {
+    ::shutdown(_connecting_socket, SHUT_RDWR);
+    ::closesocket(_connecting_socket);
+  }
 }
 
 
@@ -1475,7 +1228,8 @@ bool unix_socket_stream::isReady(unsigned int milliseconds)
 #ifdef SOCK_RAW
 
 raw_socket_stream::raw_socket_stream(FreeSockets::IP_Protocol proto) :
-    basic_socket_stream(dgram_sockbuf), dgram_sockbuf(INVALID_SOCKET)
+                  basic_socket_stream(*new dgram_socketbuf(INVALID_SOCKET)),
+                  dgram_sockbuf((dgram_socketbuf&)_sockbuf)
 {
   protocol = proto;
   SOCKET_TYPE _socket = ::socket(AF_INET, SOCK_RAW, protocol);
@@ -1484,8 +1238,11 @@ raw_socket_stream::raw_socket_stream(FreeSockets::IP_Protocol proto) :
 
 raw_socket_stream::raw_socket_stream(unsigned insize,unsigned outsize,
                                      FreeSockets::IP_Protocol proto) :
-                                 basic_socket_stream(dgram_sockbuf, proto),
-                                 dgram_sockbuf(INVALID_SOCKET,insize,outsize)
+                  basic_socket_stream(*new dgram_socketbuf(INVALID_SOCKET,
+                                                           insize,
+                                                           outsize),
+                                      proto),
+                  dgram_sockbuf((dgram_socketbuf&)_sockbuf)
 {
   protocol = proto;
   SOCKET_TYPE _socket = ::socket(AF_INET, SOCK_RAW, protocol);
@@ -1501,7 +1258,8 @@ void raw_socket_stream::setProtocol(FreeSockets::IP_Protocol proto) {
 
 raw_socket_stream::~raw_socket_stream()
 {
-    shutdown();
+  // Don't close the main socket, that is done in the basic_socket_stream
+  // destructor
 }
 
 #endif // SOCK_RAW
